#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@authors: William Woelffel & Guillaume Laurent
Generation of a noisy Raman spectrum
and denoising using Principal Component Analysis (PCA).
"""

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from sklearn import decomposition

def generate_binary(n_samples=300, n_features=800, min_content=.1, nts=.01):
    """
    Generates artificial spectra by linearly mixing two components and
    adding Gaussian noise.

    Parameters
    ----------
    n_samples : int, default 300
        Number of samples to generate
    n_features : int, default 800
        Number of points in a spectrum
    min_content : float, default .1
        Minimum amount of one component.
    nts : float, default .01
        Noise to signal ratio. Must be >= 0.

    Examples
    --------
    sigma, spectra, components, mix = generate_binary(55)
    """
    sigma = np.linspace(301., 1306., n_features)
    a = np.exp(-(sigma-450)**2 / (2*50**2))
    b = np.exp(-(sigma-900)**2 / (2*90**2))
    c = np.exp(-(sigma-510)**2 / (2*50**2))
    d = np.exp(-(sigma-750)**2 / (2*70**2))
    original_components = np.vstack((a + b / 3, c / 2 + d))
    if nts > 0:
        noise = np.random.normal(scale=nts, size=(n_samples, n_features))
    else:
        noise = np.zeros((n_samples, n_features))
    rate = np.linspace(min_content, 1 - min_content, num=n_samples)
    mix = np.column_stack((rate, 1 - rate))
    spectra = np.dot(mix, original_components) + noise
    return sigma, spectra, original_components, mix


def raman_plot(sigma, spectra, clear=False, legend=None, spectral_step=200,
        **kwargs):
    """
    Plots Raman spectra as a function of the wavenumber,
    with ticks and subticks.

    Parameters
    ----------
    sigma : numpy.ndarray, n_features
        List of the wavenumbers at which the intensities are recorded.
    spectra : numpy.ndarray, (n_spectra *) n_features
        Array containing one or more spectra to be plotted. The
        spectrum/spectra must have as many points as there are in `sigma`.
    clear : bool, default True
        Whether the plot frame should be cleared of numerical values.
    legend : iterable of n_spectra strings, optional
        The legend to display on the graph. Must have exactly as many elements
        as there are components. If None, there will be no legend.
    spectral_step : int, default 200
        Major ticks on the x axis will be placed every `spectral_step`. Minor
        ticks will be placed every `spectral_step`/5.

    Notes
    -----
    `kwargs` are passed to the `plt.plot` call.
    """
    # kwargs processing
    linewidth = kwargs.pop('lw', 2)
    # Figure setup
    mpl.style.use('classic')    # use bgrcmyk colors
    fig, ax = plt.subplots()
    sigma_major_locator = mpl.ticker.MultipleLocator(spectral_step)
    sigma_minor_locator = mpl.ticker.MultipleLocator(spectral_step / 5)
    spectrum_formatter = mpl.ticker.FormatStrFormatter('%1.1e')
    ax.xaxis.set_major_locator(sigma_major_locator)
    ax.xaxis.set_minor_locator(sigma_minor_locator)
    ax.set_xlabel('Wavenumber / cm$^{-1}$')
    ax.yaxis.set_major_formatter(spectrum_formatter)
    ax.set_ylabel('Intensity / AU')
    # Actual plotting
    lines = ax.plot(sigma, spectra.T, linewidth=linewidth, **kwargs)
    if legend and len(legend) == len(lines):
        plt.legend(lines, legend)
    elif legend is None:
        pass
    else:
        raise ValueError('Legend must have as many items as components.')
    # Postprocessing
    ax.set_xlim((sigma.min(), sigma.max()))
    ax.set_ylim((-0.2, 1.2))
    if clear is True:
        ax.yaxis.set_ticks([])
    return fig, ax


def pca_step(raw_spectra):
    """
    Performs a PCA on the input spectra, calculates the number of components
    using the Malinowski indicator function.

    Notes
    -----
    Malinowski's indicator function is explained at the following webpages
    http://www.vub.ac.be/fabi/multi/pcr/chaps/chap6.html
    http://pubs.acs.org/doi/pdf/10.1021/ac50012a027

    Parameters
    ----------
    raw_spectra : numpy.ndarray, n_spectra * n_features
        Input spectra

    Returns
    -------
    n_components : int
        Number of components from Malinowski's IND function
    denoised_spectra : numpy.ndarray, n_spectra * n_features
        Denoised spectra, reconstructed from PCA with n_components components.
    """
    n_samples, n_features = raw_spectra.shape
    pca = decomposition.PCA()
    pca.fit(raw_spectra)   # Fit the data
    eigenvalues = np.sqrt(n_samples * pca.explained_variance_)
    # Malinowski's indicator function
    real_error = np.sqrt(eigenvalues[::-1].cumsum()[::-1] /
        (n_features * (n_samples - np.arange(n_samples))))
    indicator = real_error / (n_samples - np.arange(n_samples)) **2
    n_components = 1 + np.argmin(indicator[:int(n_samples / 2)])
    print('Significant components : {:d}'.format(n_components))
    # Raw_spectra denoising
    pca.n_components = n_components
    denoised_spectra = pca.fit_transform(raw_spectra)
    denoised_spectra = pca.inverse_transform(denoised_spectra)
    return n_components, denoised_spectra


# Generate a family of n_samples artifical spectra by linarly mixing two base
# spectra, the original components, according to a mixing matrix. Gaussian
# noise is added. The noise level is scaled to be fairly important.
n_samples = 2000
n_features = 2000
sigma, spectra, o_components, o_mix = generate_binary(n_samples, n_features,
    nts=.05)

# Let us have a look at 3 of these spectra
subsampling_step = round(n_samples / 3)
fig1, ax1 = raman_plot(sigma, spectra[::subsampling_step])
fig1.show()

# Rather noisy. Let us perform PCA, select the relevant number of components to
# keep thank to Malinowski's IND function, and then drop the other components
# before reconstructing the signal.
n_components, denoised = pca_step(spectra)
print('{:d} components were kept'.format(n_components))
# Now let's see if we improved things
fig2, ax2 = raman_plot(sigma, denoised[::subsampling_step])
fig2.show()

# Finally we can compare to the true non-noisy spectra
true_spectra = np.dot(o_mix, o_components)
fig3, ax3 = raman_plot(sigma, true_spectra[::subsampling_step])
fig3.show()

input('Press Enter to exit')    # to avoid figures closing
